/* GNU readline support */

#include "bounix.h"

#ifdef WITH_READLINE

#ifdef HAVE_READLINE_H
#include <readline.h>
#endif
#ifdef HAVE_HISTORY_H
#include <history.h>
#endif
#ifdef HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif
#ifdef HAVE_READLINE_HISTORY_H
#include <readline/history.h>
#endif
#ifdef HAVE_STRINGS_H
#include <strings.h>
#endif

#define COMP_BO  1
#define COMP_NB  2

/* A static variable for holding the line. */
static char *line_read = (char *) NULL;
static int which_completion = COMP_BO;

static int my_len = -1;
static int my_idx = -1;




void sig_alrm(int signo)
{
	extern int check_netbus_response(char *buf);
	extern void nbrecognizecommand(char *buf, int len, char **c1,
				       char **c2);
	char buf[NBBUF_LEN], *c1, *c2;
	int len;

	memset(buf, 0, sizeof(buf));
	len = check_netbus_response(buf);
	if (len > 0) {
		c1 = NULL;
		c2 = NULL;
		nbrecognizecommand(buf, len, (char **) &c1, (char **) &c2);
		fprintf(rl_outstream, "\n");
		fprintf(rl_outstream, "===[ NB RESPONSE BEGIN ]===\n");
		fprintf(rl_outstream, "Command: '%s'\n", c1);
		fprintf(rl_outstream, "Info: '%s'\n", c2);
		fprintf(rl_outstream, "===[ NB RESPONSE END ]===\n");
		rl_forced_update_display();
	} else
		switch (len) {
		case NB_NO_DATA:
			break;
		case NB_SERVER_DISCONNECTED:
			fprintf(rl_outstream, "\nConnection closed by foreign host.\n\n");
			rl_forced_update_display();
			break;
		case NB_NETWORK_ERROR:
			break;
		}
	signal(signo, sig_alrm);
	alarm(ALARM);
}



static char *my_completor(char *text, int state)
{
	struct COMMAND *cmds;

	if (my_len < 0)
		state = 0;
	if (state == 0) {
		my_idx = 0;
		my_len = strlen(text);
	}
	cmds = get_commands();
	if (!cmds)
		return (char *) NULL;
	for (; cmds[my_idx].cmd != (char *) NULL; my_idx++) {
		if (strncasecmp(text, cmds[my_idx].cmd, my_len) == 0)
			return strdup(cmds[my_idx++].cmd);
	}
	return (char *) NULL;
}


static char **my_complete(char *text, int start, int end, int type)
{
	char *helpstr = "help";
	char **matches;
	int len;

	len = strlen(rl_line_buffer);
	if (len > strlen(helpstr))
		len = strlen(helpstr);
	matches = (char **) NULL;
	if (start == 0) {
		which_completion = type;
		matches = completion_matches(text, my_completor);
	}
	else if (strncasecmp(rl_line_buffer, helpstr, len) == 0) {
		which_completion = type;
		my_idx = my_len = -1;
		matches = completion_matches(text, my_completor);
	}
	return matches;
}


char **bo_complete(char *text, int start, int end)
{
	return my_complete(text, start, end, COMP_BO);
}


char **nb_complete(char *text, int start, int end)
{
	return my_complete(text, start, end, COMP_NB);
}


/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *
rl_gets(char *prompt)
{
	/* If the buffer has already been allocated, return the memory
	   to the free pool. */
	if (line_read != NULL) {
		free(line_read);
		line_read = (char *) NULL;
	}
	/* Get a line from the user. */
	line_read = readline(prompt);

	/* If the line has any text in it, save it on the history. */
	if (line_read != NULL && strlen(line_read) != 0) {
		add_history(line_read);
#ifdef DEBUG
		printf("\nlen: %d [%s] 0x%02X\n", strlen(line_read), line_read, line_read[strlen(line_read) - 1]);
#endif
	}
	return line_read;
}

#endif				/* WITH_READLINE */
