/*
 *  Copyright (C) 2000 Denis Lemire <denis@lemire.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <gtk/gtk.h>
#include <string.h>
#include "random.h"

#include "../pixmaps/exec.xpm"
#include "../pixmaps/close.xpm"

void handleCharType(GtkWidget *widget, gpointer data);
void handleAlphaCase(GtkWidget *widget, gpointer data);
void handleSingleHand(GtkWidget *widget, gpointer data);
void handlePasswdLen(GtkWidget *widget, gpointer data);
void generate(GtkWidget *widget, gpointer data);

struct passwdOptions password;

int alphaCase = 2;
int includeAlpha = TRUE;

int main(int argc, char *argv[])
{
  GtkWidget *mainWin;
  GtkWidget *vPackBox1, *vPackBox2, *vPackBox3;
  GtkWidget *hPackBox1, *hPackBox2, *hPackBox3;
  GtkWidget *hPackBox5, *hPackBox6, *hPackBox7;
  GtkWidget *table1;
  GtkWidget *frame1, *frame2, *frame3, *frame4;
  GtkWidget *label1;
  GtkWidget *entry1;
  GtkWidget *checkButton1, *checkButton2, *checkButton3;
  GtkWidget *radioButton1, *radioButton2, *radioButton3;
  GtkWidget *radioButton4, *radioButton5, *radioButton6, *radioButton7;
  GtkWidget *button1, *button2;
  GtkWidget *buttonLbl1, *buttonLbl2;
  GtkWidget *spinButton1;
  GtkWidget *pixWidget1, *pixWidget2;
  GtkWidget *hSeparator1;
  GdkPixmap *pixmap1, *pixmap2;
  GdkBitmap *mask1, *mask2;
  GdkColormap *colormap;
  GtkObject *spinButton1Adj;
  GSList *alphaCase_group = NULL;
  GSList *handSelect_group = NULL;
  char windowTitle[15];

  password.type = 7;
  password.length = 8;
  password.hand = 0;

  strcpy(windowTitle, "passwdGen ");
  strcat(windowTitle, VERSION);

  gtk_init (&argc, &argv);

  mainWin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (mainWin), windowTitle);
  gtk_window_set_policy (GTK_WINDOW (mainWin), FALSE, FALSE, FALSE);

  gtk_signal_connect (GTK_OBJECT (mainWin), "delete_event",
                      GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
  
  vPackBox1 = gtk_vbox_new(FALSE, 0);

  gtk_container_add(GTK_CONTAINER (mainWin), vPackBox1);


  /* Here starts the Generated Password Frame */
  frame1 = gtk_frame_new("Generated Password");
  hPackBox1 = gtk_hbox_new(FALSE, 0);
  entry1 = gtk_entry_new ();
  GTK_WIDGET_UNSET_FLAGS (entry1, GTK_CAN_FOCUS);
  gtk_widget_show(entry1);
  gtk_box_pack_start (GTK_BOX (hPackBox1), entry1, TRUE, TRUE, 0);
  gtk_container_set_border_width(GTK_CONTAINER (frame1), 5);
  gtk_container_set_border_width (GTK_CONTAINER (hPackBox1), 5);
  gtk_widget_show(hPackBox1);
  gtk_container_add(GTK_CONTAINER (frame1), hPackBox1);

  gtk_widget_show(frame1);

  /* Here starts the Characters Frame */
  hPackBox2 = gtk_hbox_new(FALSE, 0);
  vPackBox2 = gtk_vbox_new(FALSE, 0);
  frame2 = gtk_frame_new("Characters");

  gtk_container_set_border_width(GTK_CONTAINER (frame2), 5);

  checkButton1 = gtk_check_button_new_with_label("alphabetic");
  checkButton2 = gtk_check_button_new_with_label("numeric");
  checkButton3 = gtk_check_button_new_with_label("printable");

  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkButton1), TRUE);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (checkButton2), TRUE);

  gtk_signal_connect(GTK_OBJECT (checkButton1), "clicked",
                     GTK_SIGNAL_FUNC (handleCharType), (gpointer) 0);

  gtk_signal_connect(GTK_OBJECT (checkButton2), "clicked",
                     GTK_SIGNAL_FUNC (handleCharType), (gpointer) 1);

  gtk_signal_connect(GTK_OBJECT (checkButton3), "clicked",
                     GTK_SIGNAL_FUNC (handleCharType), (gpointer) 2);

  gtk_widget_show(checkButton1);
  gtk_widget_show(checkButton2);
  gtk_widget_show(checkButton3);

  gtk_box_pack_start(GTK_BOX (vPackBox2), checkButton1, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox2), checkButton2, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox2), checkButton3, TRUE, TRUE, 0);


  /* Here starts the Alphabet Case Frame */
  frame3 = gtk_frame_new("Alphabet Case");
  vPackBox3 = gtk_vbox_new(FALSE, 0);

  radioButton1 = gtk_radio_button_new_with_label (NULL, "lowercase");
  alphaCase_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton1));

  radioButton2 = gtk_radio_button_new_with_label (alphaCase_group, "uppercase");
  alphaCase_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton2));

  radioButton3 = gtk_radio_button_new_with_label (alphaCase_group, "mixed");
  alphaCase_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton3));

  gtk_signal_connect(GTK_OBJECT (radioButton1), "clicked",
                     GTK_SIGNAL_FUNC (handleAlphaCase), (gpointer) 0);

  gtk_signal_connect(GTK_OBJECT (radioButton2), "clicked",
                     GTK_SIGNAL_FUNC (handleAlphaCase), (gpointer) 1);

  gtk_signal_connect(GTK_OBJECT (radioButton3), "clicked",
                     GTK_SIGNAL_FUNC (handleAlphaCase), (gpointer) 2);

  gtk_widget_show(radioButton1);
  gtk_widget_show(radioButton2);
  gtk_widget_show(radioButton3);

  gtk_container_set_border_width(GTK_CONTAINER (frame3), 5);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radioButton3), TRUE);

  gtk_box_pack_start(GTK_BOX (vPackBox3), radioButton1, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox3), radioButton2, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox3), radioButton3, TRUE, TRUE, 0);

  gtk_widget_show(vPackBox2);
  gtk_widget_show(vPackBox3);
  gtk_container_add(GTK_CONTAINER (frame2), vPackBox2);
  gtk_container_add(GTK_CONTAINER (frame3), vPackBox3);
  gtk_widget_show(frame2);
  gtk_widget_show(frame3);
  gtk_box_pack_start(GTK_BOX (hPackBox2), frame2, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (hPackBox2), frame3, TRUE, TRUE, 0);
  gtk_widget_show(hPackBox2);


  /* Here starts the Single Handed Case Frame */
  frame4 = gtk_frame_new("Specific Hand");
  table1 = gtk_table_new (2, 2, TRUE);

  radioButton4 = gtk_radio_button_new_with_label (NULL, "left");
  handSelect_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton4));

  radioButton5 = gtk_radio_button_new_with_label (handSelect_group, "right");
  handSelect_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton5));

  radioButton6 = gtk_radio_button_new_with_label(handSelect_group, "alternate");
  handSelect_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton6));

  radioButton7 = gtk_radio_button_new_with_label (handSelect_group, "disabled");
  handSelect_group = gtk_radio_button_group (GTK_RADIO_BUTTON (radioButton7));

  gtk_container_set_border_width(GTK_CONTAINER (frame4), 5);
  gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radioButton7), TRUE);

  gtk_signal_connect(GTK_OBJECT (radioButton4), "clicked",
                     GTK_SIGNAL_FUNC (handleSingleHand), (gpointer) 1);

  gtk_signal_connect(GTK_OBJECT (radioButton5), "clicked",
                     GTK_SIGNAL_FUNC (handleSingleHand), (gpointer) 2);

  gtk_signal_connect(GTK_OBJECT (radioButton6), "clicked",
                     GTK_SIGNAL_FUNC (handleSingleHand), (gpointer) 3);

  gtk_signal_connect(GTK_OBJECT (radioButton7), "clicked",
                     GTK_SIGNAL_FUNC (handleSingleHand), (gpointer) 0);

  gtk_widget_show(radioButton4);
  gtk_widget_show(radioButton5);
  gtk_widget_show(radioButton6);
  gtk_widget_show(radioButton7);

  gtk_table_attach (GTK_TABLE (table1), radioButton4, 0, 1, 0, 1,
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND), 2, 0);

  gtk_table_attach (GTK_TABLE (table1), radioButton5, 0, 1, 1, 2,
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND), 2, 0);

  gtk_table_attach (GTK_TABLE (table1), radioButton6, 1, 2, 0, 1,
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND), 2, 0);

  gtk_table_attach (GTK_TABLE (table1), radioButton7, 1, 2, 1, 2,
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND),
                   (GtkAttachOptions) (GTK_FILL | GTK_EXPAND), 2, 0);

  gtk_widget_show(table1);
  gtk_widget_show(frame4);

  gtk_container_add (GTK_CONTAINER (frame4), table1);


  /* Here starts the Password Length Area */
  hPackBox3 = gtk_hbox_new(FALSE, 0);
  label1 = gtk_label_new("Password Length: ");
  spinButton1Adj = gtk_adjustment_new (8, 4, 20, 1, 1, 1);
  spinButton1 = gtk_spin_button_new (GTK_ADJUSTMENT (spinButton1Adj), 1, 0);

  gtk_misc_set_padding (GTK_MISC (label1), 5, 5);

  gtk_signal_connect(GTK_OBJECT (spinButton1), "changed",
                     GTK_SIGNAL_FUNC (handlePasswdLen), NULL);

  gtk_widget_show(label1);
  gtk_widget_show(spinButton1);

  gtk_box_pack_start(GTK_BOX (hPackBox3), label1, FALSE, FALSE, 0);
  gtk_box_pack_start(GTK_BOX (hPackBox3), spinButton1, TRUE, TRUE, 5);

  gtk_widget_show(hPackBox3);

  /* Here starts the horizontal separator */
  hSeparator1 = gtk_hseparator_new ();
  gtk_widget_show (hSeparator1);

  /* Here starts the Command Button Area */
  hPackBox5 = gtk_hbox_new(TRUE, 3);
  hPackBox6 = gtk_hbox_new(FALSE, 0);
  hPackBox7 = gtk_hbox_new(FALSE, 0);
  
  colormap = gtk_widget_get_colormap (mainWin);


  pixmap1 = gdk_pixmap_colormap_create_from_xpm_d ( NULL, colormap, &mask1,
                                                    NULL, (gchar **)exec_xpm );

  pixmap2 = gdk_pixmap_colormap_create_from_xpm_d ( NULL, colormap, &mask2,
                                                    NULL, (gchar **)close_xpm );

  pixWidget1 = gtk_pixmap_new (pixmap1, mask1);
  pixWidget2 = gtk_pixmap_new (pixmap2, mask2);

  gtk_widget_show (pixWidget1);
  gtk_widget_show (pixWidget2);

  button1 = gtk_button_new();
  button2 = gtk_button_new();

  buttonLbl1 = gtk_label_new ("Generate");
  buttonLbl2 = gtk_label_new ("Close");

  gtk_misc_set_alignment(GTK_MISC(buttonLbl1), 0.5f, 0.5f);
  gtk_misc_set_alignment(GTK_MISC(buttonLbl2), 0.5f, 0.5f);

  gtk_widget_show (buttonLbl1);
  gtk_widget_show (buttonLbl2);

  gtk_box_pack_start(GTK_BOX (hPackBox6), pixWidget1, FALSE, FALSE, 3);
  gtk_box_pack_start(GTK_BOX (hPackBox7), pixWidget2, FALSE, FALSE, 3);
  gtk_box_pack_start(GTK_BOX (hPackBox6), buttonLbl1, TRUE, TRUE, 3);
  gtk_box_pack_start(GTK_BOX (hPackBox7), buttonLbl2, TRUE, TRUE, 3);

  gtk_container_set_border_width (GTK_CONTAINER (hPackBox6), 3);
  gtk_container_set_border_width (GTK_CONTAINER (hPackBox7), 3);

  gtk_widget_show (hPackBox6);
  gtk_widget_show (hPackBox7);

  gtk_container_add(GTK_CONTAINER (button1), hPackBox6);
  gtk_container_add(GTK_CONTAINER (button2), hPackBox7);

  gtk_signal_connect(GTK_OBJECT (button1), "clicked",
                     GTK_SIGNAL_FUNC (generate), GTK_OBJECT (entry1));

  gtk_signal_connect(GTK_OBJECT (button2), "clicked",
                     GTK_SIGNAL_FUNC (gtk_main_quit), NULL);

  gtk_widget_show(button1);
  gtk_widget_show(button2);

  gtk_box_pack_start(GTK_BOX (hPackBox5), button1, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (hPackBox5), button2, TRUE, TRUE, 0);

  gtk_container_set_border_width (GTK_CONTAINER (hPackBox5), 3);

  gtk_widget_show(hPackBox5); 

  gtk_box_pack_start(GTK_BOX (vPackBox1), frame1, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox1), hPackBox2, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox1), frame4, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox1), hPackBox3, TRUE, TRUE, 3);
  gtk_box_pack_start(GTK_BOX (vPackBox1), hSeparator1, TRUE, TRUE, 0);
  gtk_box_pack_start(GTK_BOX (vPackBox1), hPackBox5, TRUE, TRUE, 0);

  gtk_widget_show(vPackBox1);
  gtk_widget_show(mainWin);

  gtk_widget_grab_focus(button1);
  
  gtk_main();

  return(0);
}

void handleCharType(GtkWidget *widget, gpointer data)
{
  switch (GPOINTER_TO_INT(data)) {
    case 0:
      switch (alphaCase) {
        case 0:
          if (GTK_TOGGLE_BUTTON (widget)->active) {
            password.type = password.type + 1;
            includeAlpha = TRUE;
          } else {
            password.type = password.type - 1;
            includeAlpha = FALSE;
          }
          break;

        case 1:
          if (GTK_TOGGLE_BUTTON (widget)->active) {
            password.type = password.type + 2;
            includeAlpha = TRUE;
          } else {
            password.type = password.type - 2;
            includeAlpha = FALSE;
          }
          break;

        case 2:
          if (GTK_TOGGLE_BUTTON (widget)->active) {
            password.type = password.type + 3;
            includeAlpha = TRUE;
          } else {
            password.type = password.type - 3;
            includeAlpha = FALSE;
          }
          break;
      }
      break;

    case 1:
      if (GTK_TOGGLE_BUTTON (widget)->active) {
        password.type = password.type + 4;
      } else {
        password.type = password.type - 4;
      }
      break;

    case 2:
      if (GTK_TOGGLE_BUTTON (widget)->active) {
        password.type = password.type + 8;
      } else {
        password.type = password.type - 8;
      }
      break;
  }
}

void handleAlphaCase(GtkWidget *widget, gpointer data)
{
  switch (GPOINTER_TO_INT (data)) {
    case 0: /* Lowercase radio button */
      switch (alphaCase) {
        case 1: /* Was previously uppercase */
          alphaCase = 0;
          if (includeAlpha == TRUE)
            password.type = password.type - 1;
          break;

        case 2: /* Was previously mixed case */
          alphaCase = 0;
          if (includeAlpha == TRUE)
            password.type = password.type - 2;
          break;
      }
      break;

    case 1: /* Uppercase radio button */
      switch (alphaCase) {
        case 0: /* Was previously lowercase */
          alphaCase = 1;
          if (includeAlpha == TRUE)
            password.type = password.type + 1;
          break;

        case 2: /* Was previously mixed case */
          alphaCase = 1;
          if (includeAlpha == TRUE)
            password.type = password.type - 1;
          break;
      }
    break;

    case 2: /* Mixed case radio button */
      switch (alphaCase) {
        case 0: /* Was previously lowercase */
          alphaCase = 2;
          if (includeAlpha == TRUE)
            password.type = password.type + 2;
          break;

        case 1: /* Was previously uppercase */
          alphaCase = 2;
          if (includeAlpha == TRUE)
            password.type = password.type + 1;
          break;
      }
    break;
  }
}

void handleSingleHand(GtkWidget *widget, gpointer data) {
  password.hand = GPOINTER_TO_INT (data);
}

void handlePasswdLen(GtkWidget * widget, gpointer data)
{
  password.length = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON (widget));
}

void generate(GtkWidget *widget, gpointer data)
{
  char randPass[50];
  buildPasswd(password.type, password.length, password.hand, &randPass[0]);
  gtk_entry_set_text (GTK_ENTRY (data), randPass);
}
