/*
 *  Copyright (C) 2000 Denis Lemire <denis@lemire.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

static int seedRandom(void) {
  FILE *devRandom;

  if ((devRandom = fopen("/dev/random", "r")) != NULL) {
    srand((getc(devRandom)<<24)|(getc(devRandom)<<16)|
          (getc(devRandom)<<8)|getc(devRandom));
    fflush(devRandom);
    return (0);
  } else {
    srand((unsigned int)time((time_t *)NULL)); /* Can you say obsfucated? */
    return (1);
  }
}

static char getASCII(int passwdType) {
  char lowerCase[26] = "abcdefghijklmnopqrstuvwxyz";
  char upperCase[26] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
  char numeric[10] = "1234567890";
  char printAble[18] = "!@#$%^&*-_=+,<.>/?";
  int tmpVar;

  switch (passwdType) {
    case 1: /* Lowercase alphabetic */
    return (lowerCase[rand()%26]);

    case 2: /* Uppercase alphabetic */
    return (upperCase[rand()%26]);

    case 3: /* Mixed case alphabetic */ 
    tmpVar = rand()%52;
    if (tmpVar > 25) {
      return (upperCase[tmpVar - 26]);
    } else {
      return (lowerCase[tmpVar]);
    }

    case 4: /* Numeric */
    return (numeric[rand()%10]);

    case 5: /* Lowercase alphanumeric */
    tmpVar = rand()%36;
    if (tmpVar > 9) {
      return (lowerCase[tmpVar - 10]);
    } else {
      return (numeric[tmpVar]);
    }

    case 6: /* Uppercase alphanumeric */
    tmpVar = rand()%36;
    if (tmpVar > 9) {
      return (upperCase[tmpVar - 10]);
    } else {
      return (numeric[tmpVar]);
    }

    case 7: /* Mixed case alphanumeric */
    tmpVar = rand()%62;
    if (tmpVar > 9) {
      if (tmpVar > 35) {
        return (upperCase[tmpVar - 36]); 
      } else {
        return (lowerCase[tmpVar - 10]);
      }
    } else {
      return (numeric[tmpVar]);
    }

    case 8: /* Nonalphanumeric printable */
    return (printAble[rand()%18]);

    case 9: /* Lowercase alphabetic and printable */
    tmpVar = rand()%44;
    if (tmpVar > 17) {
      return (lowerCase[tmpVar - 18]);
    } else {
      return (printAble[tmpVar]);
    }

    case 10: /* Uppercase alphabetic and printable */
    tmpVar = rand()%44;
    if (tmpVar > 17) {
      return (upperCase[tmpVar - 18]);
    } else {
      return (printAble[tmpVar]);
    }

    case 11: /* Mixed case alphabetic and printable */
    tmpVar = rand()%62;
    if (tmpVar > 17) {
      if (tmpVar > 43) {
        return (upperCase[tmpVar - 44]); 
      } else {
        return (lowerCase[tmpVar - 18]);
      }
    } else {
      return (printAble[tmpVar]);
    }

    case 12: /* Numeric and printable */
    tmpVar = rand()%28;
    if (tmpVar > 17) {
      return (numeric[tmpVar - 18]);
    } else {
      return (printAble[tmpVar]);
    }

    case 13: /* Lowercase alphanumeric and printable */
    tmpVar = rand()%54;
    if (tmpVar > 17) {
      if (tmpVar > 27) {
        return (lowerCase[tmpVar - 28]);
      } else {
        return (numeric[tmpVar - 18]);
      }
    } else {
      return (printAble[tmpVar]);
    }

    case 14: /* Uppercase alphanumeric and printable */
    tmpVar = rand()%54;
    if (tmpVar > 17) {
      if (tmpVar > 27) {
        return (upperCase[tmpVar - 28]);
      } else {
        return (numeric[tmpVar - 18]);
      }
    } else {
      return (printAble[tmpVar]);
    }

    case 15: /* Mixed case alphanumeric and printable */
    tmpVar = rand()%80;
    if (tmpVar > 17) {
      if (tmpVar > 27) {
        if (tmpVar > 53) {
          return (upperCase[tmpVar - 54]);
        } else {
          return (lowerCase[tmpVar - 28]);
        }
      } else {
        return (numeric[tmpVar - 18]);
      }
    } else {
      return (printAble[tmpVar]);
    }
  }
  return ('\0');
}

int buildPasswd(int passwdType, int passwdLen, int hand, char *pwString_ptr) {
  char *randPasswd_ptr[passwdLen + 1];
  char leftKeys[42] = "!#$%123456@ABCDEFGQRSTVWXZabcdefgqrstvwxz^";
  char rightKeys[40] = "&*+,_./0789<=>?HIJKLMNOPUYhijklmnopuy[]-";
  char tmpChar;
  int seedStatus;
  int handOK = 0;
  int counter[2];

  seedStatus = seedRandom();

  randPasswd_ptr[0] = pwString_ptr;
  for (counter[0]=1; counter[0]<=passwdLen; counter[0]++) {
     randPasswd_ptr[counter[0]] = (pwString_ptr+counter[0]);
  } 

  switch (hand) {
    case 0: /* both hands */
      for (counter[0]=0; counter[0]<=passwdLen; counter[0]++) {
        *randPasswd_ptr[counter[0]] = getASCII(passwdType);
      }
      break;

    case 1: /* left hand */
      for (counter[0]=0; counter[0]<=passwdLen; counter[0]++) {
        do {
          if (passwdType == 0) handOK = 1;
          tmpChar = getASCII(passwdType);
          for (counter[1]=0; counter[1]<=40; counter[1]++) {
            if (tmpChar == leftKeys[counter[1]]) {
              handOK = 1; 
            }
          }
        } while (handOK != 1);
        handOK = 0;
        *randPasswd_ptr[counter[0]] = tmpChar;
      }
      break;

    case 2: /* right hand */
      for (counter[0]=0; counter[0]<=passwdLen; counter[0]++) {
        do {
          if (passwdType == 0) handOK = 1;
          tmpChar = getASCII(passwdType);
          for (counter[1]=0; counter[1]<=38; counter[1]++) {
            if (tmpChar == rightKeys[counter[1]]) {
              handOK = 1; 
            }
          }
        } while (handOK != 1);
        handOK = 0;
        *randPasswd_ptr[counter[0]] = tmpChar;
      }
      break;

    case 3: /* alternating hands :) */
      for (counter[0]=0; counter[0]<=passwdLen; counter[0]++) {
        do {
          if (passwdType == 0) handOK = 1;
          tmpChar = getASCII(passwdType);
          /* when counter is even get a right char; odd: left char */
          if (counter[0]%2) { 
            for (counter[1]=0; counter[1]<=38; counter[1]++) {
              if (tmpChar == rightKeys[counter[1]]) {
                handOK = 1; 
              }
            }
          } else {
            for (counter[1]=0; counter[1]<=38; counter[1]++) {
              if (tmpChar == leftKeys[counter[1]]) {
                handOK = 1; 
              }
            }
          }
	} while (handOK != 1);
        handOK = 0;
        *randPasswd_ptr[counter[0]] = tmpChar;
      }
    break;
  }

  *randPasswd_ptr[passwdLen] = '\0';
  return (seedStatus);
}
